import ipaddress
from uuid import uuid4

from orchestrator.forms import FormPage
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State
from orchestrator.types import SubscriptionLifecycle, UUIDstr
from orchestrator.workflow import done, init, step, workflow
from orchestrator.workflows.steps import resync, set_status
from orchestrator.workflows.steps import store_process_subscription
from orchestrator.workflows.utils import wrap_create_initial_input_form

from gso.products.product_types.device import DeviceVendor, DeviceInactive, \
    DeviceProvisioning
from gso.services import provisioning_proxy
from gso.services.provisioning_proxy import confirm_pp_results, \
    await_pp_results


def initial_input_form_generator(product_name: str) -> FormGenerator:
    class CreateDeviceForm(FormPage):
        class Config:
            title = product_name

        fqdn: str
        ts_address: ipaddress.IPv4Address
        ts_port: int
        device_vendor: DeviceVendor

    user_input = yield CreateDeviceForm

    return user_input.dict()


@step('Create subscription')
def create_subscription(product: UUIDstr) -> State:
    subscription = DeviceInactive.from_product_id(product, uuid4())

    return {
        'subscription': subscription,
        'subscription_id': subscription.subscription_id,
    }


@step('Get information from IPAM')
def get_info_from_ipam(subscription: DeviceInactive) -> State:
    # lo = ipam.new_device_lo_address()
    # subscription.device.lo_ipv4_address = lo.v4
    # subscription.device.lo_ipv6_address = lo.v6
    # TODO: get info about how these should be generated
    subscription.device.lo_ipv4_address = '10.10.10.10'
    subscription.device.lo_ipv6_address = 'fc00:798:10::10'
    subscription.device.lo_iso_address = '49.51e5.0001.0620.4009.6047.00'
    subscription.device.si_ipv4_network = '192.168.0.0/31'
    subscription.device.ias_lt_ipv4_network = '192.168.1.0/31'
    subscription.device.ias_lt_ipv6_network = 'fc00:798:1::150/126'

    return {'subscription': subscription}


@step('Get information about SNMP')
def get_snmp_info(subscription: DeviceInactive) -> State:
    country = 'Spain'
    city = 'Barcelona'
    country_code = 'ES'
    latitude = '41.3743'
    longitude = '2.1328'
    subscription.device.site_country = country
    subscription.device.site_city = city
    subscription.device.site_country_code = country_code
    subscription.device.site_latitude = latitude
    subscription.device.site_longitude = longitude
    subscription.device.snmp_location = (
        f'{city.upper()},{country.upper()}[{latitude},{longitude}]'
    )

    return {'subscription': subscription}


@step('Initialize subscription')
def initialize_subscription(
        subscription: DeviceInactive,
        fqdn: str,
        ts_address: ipaddress.IPv4Address,
        ts_port: str,
        device_vendor: DeviceVendor
) -> State:
    subscription.device.fqdn = fqdn
    subscription.device.ts_address = str(ts_address)
    subscription.device.ts_port = str(ts_port)
    subscription.device_vendor = device_vendor
    subscription.description = f'Device {fqdn} ' \
                               f'({subscription.device_type})'
    subscription = DeviceProvisioning.from_other_lifecycle(
        subscription, SubscriptionLifecycle.PROVISIONING
    )

    return {'subscription': subscription}


@step('Provision device [DRY RUN]')
def provision_device_dry(subscription: DeviceProvisioning,
                         process_id: UUIDstr) -> State:
    provisioning_proxy.provision_node(
        subscription,
        process_id
    )

    return {'subscription': subscription}


@step('Provision device [FOR REAL]')
def provision_device_real(subscription: DeviceProvisioning,
                          process_id: UUIDstr) -> State:
    provisioning_proxy.provision_node(
        subscription,
        process_id,
        False  # No dry run this time, run it for real
    )

    return {'subscription': subscription}


@workflow(
    'Create device',
    initial_input_form=wrap_create_initial_input_form(
        initial_input_form_generator),
    target=Target.CREATE,
)
def create_device():
    return (
            init
            >> create_subscription
            >> store_process_subscription(Target.CREATE)
            >> get_info_from_ipam
            >> get_snmp_info
            >> initialize_subscription
            >> provision_device_dry
            >> await_pp_results
            >> confirm_pp_results
            >> provision_device_real
            >> await_pp_results
            >> confirm_pp_results
            >> set_status(SubscriptionLifecycle.ACTIVE)
            >> resync
            >> done
    )
